import os
import logging
import re

from snapstraction.util import ADDR_RE

log = logging.getLogger(__name__)

LINE_RE = re.compile(r'^\S+=[0-9A-Fa-f]{6}$')

RESERVED_NAMES = ['all', 'bridge']


def _env_override():
    """Try to find SNAP_ALIASES env var"""

    try:
        return os.environ['SNAP_ALIASES']
    except:
        return None


def user_path():
    return _env_override() or os.path.expanduser(os.path.join('~', '.snap', 'aliases'))


class Aliases(object):
    """Simple cache of aliases"""

    def __init__(self):
        self.ALIAS_FILE = user_path()
        self.aliases = {}

    def _ensure_file_exists(self):
        with open(self.ALIAS_FILE, 'a'):
            pass

    def load(self):
        self._ensure_file_exists()

        with open(self.ALIAS_FILE, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if LINE_RE.match(line):
                    alias, addr = line.split('=', 1)
                    alias = alias.lower()
                    if alias in RESERVED_NAMES:
                        continue
                    self.aliases[alias] = addr.lower()

        log.debug("Node Aliases: %r", self.aliases)

    def save(self):
        with open(self.ALIAS_FILE, 'w') as f:
            for alias, addr in sorted(self.aliases.iteritems()):
                alias = alias.lower()
                if alias in RESERVED_NAMES:
                    continue
                f.write("{0}={1}\n".format(alias, addr.lower()))

    def __contains__(self, item):
        return item in self.aliases

    def __len__(self):
        return len(self.aliases)

    def __iter__(self):
        return self.aliases.__iter__()

    def __getitem__(self, item):
        return self.aliases[item]

    def add(self, alias, address):
        alias = alias.lower()
        if alias in RESERVED_NAMES:
            log.error("Cannot alias reserved name %r", alias)
            return

        if not ADDR_RE.match(address):
            log.error("Address %r is malformed, must be valid 6-char address", address)
            return

        self.aliases[alias] = address

    def discard(self, alias):
        alias = alias.lower()
        return self.aliases.pop(alias, None)

    def clear(self):
        self.aliases = {}

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()

    def translate(self, address_or_alias):
        """If the given arg is an alias we know, return the address.

        Otherwise, return the arg unchanged.
        """
        return self.aliases.get(address_or_alias.lower(), address_or_alias)
