"""SNAPtoolbelt Configuration Commands"""
import logging
import sys
from textwrap import dedent
from argparse import RawDescriptionHelpFormatter

from arghandler import ArgumentHandler
import progressbar

from snap_profile.network import Network
from snap_profile.profile import Profile, SerialProfile, TcpProfile, MqttProfile

import snapstraction.probe
from snapstraction.node_cache import NodeCache
from snapstraction.util import ADDR_RE, parse_value_for_snap

from .alias import Aliases, RESERVED_NAMES as RESERVED_ALIASES

log = logging.getLogger(__name__)

RESERVED_NAMES = ['all']


def _check_reserved_names(name):
    """Check whether name is reserved, emit error and exit if it is."""
    if name in RESERVED_NAMES:
        log.error(
            "%s is a reserved name. (Other reserved names: %s)",
            name,
            ', '.join(RESERVED_NAMES),
        )
        sys.exit(1)


def config(parser, context, args):
    """Configuration-related commands"""
    handler = ArgumentHandler(
        usage="snap [global options] config (new | copy | delete | list | move | scan | profile | network | node-cache | alias | unalias) [options]",
        description=dedent("""
        SNAPtoolbelt Configuration Commands:
          new (profile | network): Create a new profile or network
          copy (profile | network): Copy a profile or network
          delete (profile | network): Delete a profile or network
          list (profile | network): List profiles or networks
          move (profile | network): Move a profile or network
          scan : Scan for new serial profiles
          profile: View/Edit a profile
          network: View/Edit a network
          node-cache: Show/Clear the node cache
          alias: Define an alias for a node address
          unalias: Remove a previously set alias
        """),
        formatter_class=RawDescriptionHelpFormatter,
    )

    handler.set_subcommands({
        'new': config_new,
        'copy': config_copy,
        'delete': config_delete,
        'list': config_list,
        'move': config_move,
        'scan': probe,
        'profile': profile_gitalike,
        'network': network_gitalike,
        'node-cache': node_cache,
        'alias': get_alias,
        'unalias': unalias
    })

    handler.run(args, context_fxn=lambda x: context)


def config_new(parser, context, args):
    """New Configuration-related commands"""
    handler = ArgumentHandler(
        usage="snap [global options] config new [-h] (profile | network) [options]",
        description="Create a new profile or network",
    )

    handler.set_subcommands({
        'profile': new_profile,
        'network': new_network,
    })

    handler.run(args, context_fxn=lambda x: context)


def config_delete(parser, context, args):
    """Delete Configuration-related commands"""
    handler = ArgumentHandler(
        usage="snap [global options] config delete [-h] (profile | network) <name>",
        description="Delete a profile or network",
    )

    handler.set_subcommands({
        'profile': delete_profile,
        'network': delete_network,
    })

    handler.run(args, context_fxn=lambda x: context)


def config_copy(parser, context, args):
    """Copy Configuration-related commands"""
    handler = ArgumentHandler(
        usage="snap [global options] config copy [-h] (profile | network) <from> <to>",
        description="Copy a profile or network",
    )

    handler.set_subcommands({
        'profile': copy_profile,
        'network': copy_network,
    })

    handler.run(args, context_fxn=lambda x: context)


def config_move(parser, context, args):
    """Copy Configuration-related commands"""
    handler = ArgumentHandler(
        usage="snap [global options] config move [-h] (profile | network) <from> <to>",
        description="Move a profile or network",
    )

    handler.set_subcommands({
        'profile': move_profile,
        'network': move_network,
    })

    handler.run(args, context_fxn=lambda x: context)


def config_list(parser, context, args):
    """Listing Configuration-related commands"""
    handler = ArgumentHandler(
        usage="snap [global options] config list [-h] (profile | network)",
        description="List profiles or networks",
    )

    handler.set_subcommands({
        'profile': list_profile,
        'network': list_network,
    })

    handler.run(args, context_fxn=lambda x: context)


def new_profile(parser, context, args):
    """Create a new profile"""

    handler = ArgumentHandler(
        usage="snap [global options] config new profile [-h] <kind> <name> [-f]",
        description="Create a new profile",
    )

    handler.add_argument(
        "kind",
        type=str.lower,
        choices=("serial", "tcp", "mqtt"),
        metavar="KIND",
        help="Kind of profile to create (serial, tcp, mqtt)",
    )
    handler.add_argument(
        "name",
        metavar="NAME",
        help="Name of profile",
    )
    handler.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Overwrite the destination profile, if it exists",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config new profile command: %r", rest)
        sys.exit(1)

    _check_reserved_names(args.name)

    log.info("Creating a new %s profile named %s as a copy of default",
             args.kind, args.name)

    # UGLY!
    # TODO: replace with Profile __new__ factory?
    if args.kind == "serial":
        profile = SerialProfile()
    elif args.kind == "tcp":
        profile = TcpProfile()
    elif args.kind == "mqtt":
        profile = MqttProfile()

    if args.name in Profile.list():
        if args.force:
            log.warn("Destination profile %s already exists, overwriting",
                     args.name)
            Profile.delete(args.name)
        else:
            log.error("Destination profile %s already exists, aborting!",
                      args.name)
            sys.exit(1)

    # Get the new name
    profile.name = args.name
    log.debug("New profile: %r", profile)
    profile.save()


def delete_profile(parser, context, args):
    """Delete a profile"""

    handler = ArgumentHandler(
        usage="snap [global options] config delete profile [-h] <name>",
        description="Delete a profile",
    )
    handler.add_argument('name')

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config delete profile command: %r", rest)
        sys.exit(1)

    log.info("Deleting profile %s", args.name)

    if args.name == "all":
        for name in Profile.list():
            Profile.delete(name)

    else:
        Profile.delete(args.name)


def copy_profile(parser, context, args):
    """Copy a profile"""

    handler = ArgumentHandler(
        usage="snap [global options] config copy profile [-h] <from_profile> <to_profile>",
        description="Copy a profile",
    )
    handler.add_argument("from_profile")
    handler.add_argument("to_profile")
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config copy profile command: %r", rest)
        sys.exit(1)

    _check_reserved_names(args.to_profile)

    if args.from_profile == args.to_profile:
        log.error("Cannot copy profile %s to itself.", args.from_profile)
        sys.exit(1)

    log.info("Copying profile %s to %s", args.from_profile, args.to_profile)

    profile = Profile.load(args.from_profile)

    # Get the new name
    profile.name = args.to_profile
    log.debug("To profile: %r", profile)
    profile.save()


def move_profile(parser, context, args):
    """Move a profile"""

    handler = ArgumentHandler(
        usage="snap [global options] config move profile [-h] <from_profile> <to_profile>",
        description="Move a profile",
    )
    handler.add_argument("from_profile")
    handler.add_argument("to_profile")
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config move profile command: %r", rest)
        sys.exit(1)

    _check_reserved_names(args.to_profile)

    if args.from_profile == args.to_profile:
        log.error("Cannot move profile %s to itself.", args.from_profile)
        sys.exit(1)

    log.info("Renaming profile %s to %s", args.from_profile, args.to_profile)

    profile = Profile.load(args.from_profile)

    # Get the new name
    profile.name = args.to_profile
    log.debug("To profile: %r", profile)
    profile.save()

    # Now remove the old one.
    Profile.delete(args.from_profile)


def list_profile(parser, context, args):
    """List all the available profiles"""

    handler = ArgumentHandler(
        usage="snap [global options] config list profile [-h]",
        description="List profiles",
    )
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config list profile command: %r", rest)
        sys.exit(1)

    log.info("Listing profiles")

    map(context.output, Profile.list())


def profile_gitalike(parser, context, args):
    """git config-like interface for profiles"""

    handler = ArgumentHandler(
        usage="snap [global options] config profile <profile_name> [<key> [<value>]]",
        description=dedent("""
        Query/Set Profile Parameters:
          (no key): Show all profile parameters
          <key>: Show parameter by key
          <key> <value>: Set parameter to value
        """),
        formatter_class=RawDescriptionHelpFormatter,
    )

    handler.add_argument("profile_name")
    handler.add_argument("key", nargs='?', default=None)
    handler.add_argument("value", nargs='?', default=None)

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config profile command: %r", rest)
        sys.exit(1)

    log.info("profile config")
    log.info("ARGS: %r", args)

    # Try to read the profile
    profile = Profile.load(args.profile_name)

    log.info("Profile: %r", profile)

    if args.profile_name != profile.name:
        log.warn("Profile %s does not exist, %s was loaded", args.profile_name,
                 profile.name)

    if args.key is None:
        log.debug("Displaying profile: %s", profile.name)
        context.output(vars(profile))
    elif args.value is None:
        log.debug("Displaying %s in profile: %s", args.key, profile.name)
        context.output({args.key: getattr(profile, args.key)})
    else:
        log.debug("Updating %s = %s in profile: %s", args.key, args.value,
                  profile.name)

        if args.key in ("speed"):
            args.value = int(args.value)

        setattr(profile, args.key, args.value)
        context.output({args.key: getattr(profile, args.key)})
        profile.save()


def new_network(parser, context, args):
    """Create a network"""

    handler = ArgumentHandler(
        usage="snap [global options] config new network [-h] <kind> <name> [-f]",
        description="Create a new network",
    )
    handler.add_argument("name")
    handler.add_argument("-f", "--force", action="store_true")
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config new network command: %r", rest)
        sys.exit(1)

    log.info("Creating a new network named %s as a copy of default",
             args.name)

    _check_reserved_names(args.name)

    # Clone the default
    net = Network.load('default')

    if args.name in Network.list():
        if args.force:
            log.warn("Destination network %s already exists, overwriting",
                     args.name)
        else:
            log.error("Destination network %s already exists, aborting!",
                      args.name)
            sys.exit(1)

    net.name = args.name
    log.debug("New network: %r", net)
    net.save()


def list_network(parser, context, args):
    """List networks"""

    handler = ArgumentHandler(
        usage="snap [global options] config list network [-h]",
        description="List all networks",
    )
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config list network command: %r", rest)
        sys.exit(1)

    log.info("Listing networks")

    map(context.output, Network.list())


def delete_network(parser, context, args):
    """Delete a network"""

    handler = ArgumentHandler(
        usage="snap [global options] config delete network [-h] <name>",
        description="Delete a network",
    )
    handler.add_argument('network_name')

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config network delete command: %r", rest)
        sys.exit(1)

    log.info("Deleting network %s", args.network_name)

    if args.network_name == "all":
        for name in Network.list():
            log.debug("deleting network %r", name)
            Network.delete(name)
    else:
        Network.delete(args.network_name)


def copy_network(parser, context, args):
    """Copy a network"""

    handler = ArgumentHandler(
        usage="snap [global options] config copy network [-h] <from_network> <to_network>",
        description="Copy a network",
    )
    handler.add_argument('from_network')
    handler.add_argument('to_network')
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config network copy command: %r", rest)
        sys.exit(1)

    _check_reserved_names(args.to_network)

    if args.from_network == args.to_network:
        log.error("Cannot copy network %s to itself.", args.from_network)
        sys.exit(1)

    log.info("Copying network from %s to %s", args.from_network,
             args.to_network)

    net = Network.load(args.from_network)

    # Get the new name
    net.name = args.to_network
    log.debug("To network: %r", net)
    net.save()


def move_network(parser, context, args):
    """Move a network"""

    handler = ArgumentHandler(
        usage="snap [global options] config move network [-h] <from_network> <to_network>",
        description="Move a network",
    )
    handler.add_argument('from_network')
    handler.add_argument('to_network')
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config network move command: %r", rest)
        sys.exit(1)

    _check_reserved_names(args.to_network)

    if args.from_network == args.to_network:
        log.error("Cannot move network %s to itself.", args.from_network)
        sys.exit(1)

    log.info("Renaming network from %s to %s", args.from_network,
             args.to_network)

    net = Network.load(args.from_network)

    # Get the new name
    net.name = args.to_network
    log.debug("To network: %r", net)
    net.save()

    Network.delete(args.from_network)


def network_gitalike(parser, context, args):
    """git config-like interface for networks"""

    handler = ArgumentHandler(
        usage="snap [global options] config network <network_name> [<key> [<value>]]",
        description=dedent("""
        Query/Set Network Parameters:
          (no key): Show all network parameters
          <key>: Show parameter by key
          <key> <value>: Set parameter to value
        """),
        formatter_class=RawDescriptionHelpFormatter,
    )

    handler.add_argument("network_name")
    handler.add_argument("key", nargs='?', default=None)
    handler.add_argument("value", nargs='?', default=None)

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config network command: %r", rest)
        sys.exit(1)

    log.info("network config")

    # Try to read the network
    network = Network.load(args.network_name)

    if args.network_name != network.name:
        log.warn("Network %s does not exist, %s was loaded", args.network_name,
                 network.name)

    if args.key is None:
        log.debug("Displaying network %s", network.name)
        output = vars(network)
    elif args.value is None:
        log.debug("Displaying %s in network %s", args.key, network.name)
        output = {args.key: getattr(network, args.key)}
    else:
        args.value = parse_value_for_snap(args.value)
        log.debug("Updating %s = %s in network %s", args.key, args.value,
                  network.name)
        log.info("Updating network")

        if args.key == "net_id":
            # Format as string to satisfy snap-profile...
            args.value = "0x%04x" % args.value

        if args.key in ("encryption_type"):
            # Fix up "None" as a valid encryption type...
            if args.value is None:
                args.value = "None"

        if args.key in ("second_crc", "packet_crc"):
            # Fix up Booleans...
            args.value = bool(args.value)

        setattr(network, args.key, args.value)
        network.save()
        output = {args.key: getattr(network, args.key)}

    # String-Escape encryption_key to deal sanely with unprintable chars.
    if "encryption_key" in output:
        output["encryption_key"] = output["encryption_key"].encode("string_escape")

    context.output(output)


def probe(parser, context, args):  # noqa
    """Probe for local SNAP connections, output on stdout as profile block"""

    handler = ArgumentHandler(
        usage="snap [global options] config scan [-h] [-d] [-f]",
        description="Probe for new serial profiles",
    )

    handler.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        default=False,
        help="Just scan, don't save new profiles",
    )

    handler.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Ignore read errors in result",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config scan command: %r", rest)
        sys.exit(1)

    log.info("Probing for local SNAP nodes")

    widgets = [progressbar.FormatLabel('Scanning for nodes:'), ' ',
               progressbar.BouncingBar(marker=progressbar.RotatingMarker()),
               ' ', progressbar.Timer()]

    progress = progressbar.ProgressBar(widgets=widgets)
    progress.start()

    def on_progress(kind, count):
        """Update progress bar"""
        widgets[0] = progressbar.FormatLabel('Scanning for %s nodes:' % kind)
        progress.update(count)

    def on_finished():
        """Finish progress bar"""
        progress.finish()

    new_profiles = snapstraction.probe.probe_serial_nodes(
        force=args.force,
        dry_run=args.dry_run,
        timeout=context.timeout,
        progress_cb=on_progress,
        finished_cb=on_finished)

    log.info("Got new profiles: %r", new_profiles)
    if new_profiles:
        map(context.output, new_profiles)


def node_cache(parser, context, args):
    """Node Cache-related commands"""
    handler = ArgumentHandler(
        usage="snap [global options] config node-cache [-h] (clear | show) [options]",
        description=dedent("""
        Node Cache Commands:
          clear: Remove nodes from the cache
          show: Show the cache
        """),
        formatter_class=RawDescriptionHelpFormatter,
    )

    # Add subcommands here.
    handler.set_subcommands({
        'clear': node_cache_clear,
        'show': node_cache_show,
    })

    handler.run(args, context_fxn=lambda x: context)


def node_cache_clear(parser, context, args):
    """Remove one or all elements from the node cache"""

    handler = ArgumentHandler(
        usage="snap [global options] config node-cache clear [-h] (all | <address>)",
        description="Remove the specified address (or all nodes) from the cache",
    )
    handler.add_argument("address", help="address to remove or 'all'")
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config node-cache clear command: %r", rest)
        sys.exit(1)

    if args.address.lower() == 'all':
        with NodeCache() as cache:
            cache.clear()

    else:
        with NodeCache() as cache:
            cache.discard(args.address)


def node_cache_show(parser, context, args):
    """Show all elements from the node cache"""

    handler = ArgumentHandler(
        usage="snap [global options] config node-cache show [-h]",
        description="Show the nodes in the cache",
    )
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config node-cache show command: %r", rest)
        sys.exit(1)

    with NodeCache() as cache:
        for addr in sorted(cache):
            context.output(addr)


def get_alias(parser, context, args):
    """List or define aliases for SNAP node addresses"""
    handler = ArgumentHandler(
        usage="snap [global options] config alias [-h] [<alias>[=<target>]]",
        description="List or define aliases for SNAP node addresses",
    )
    handler.add_argument("alias", nargs='?', default=None)
    handler.add_argument("target", nargs='?', default=None)
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config alias command: %r", rest)
        sys.exit(1)

    log.debug("Aliasing: %r", args.alias)

    if args.alias:
        # Looking up a single alias or defining a new one
        if '=' not in args.alias and not args.target:
            # Assume you're trying to look up an alias
            with Aliases() as aliases:
                args.alias = args.alias.lower()
                if args.alias in aliases:
                    context.output({args.alias: aliases[args.alias]})
                else:
                    log.error("Alias %r is not defined", args.alias)
        else:
            # Defining a new alias
            if '=' in args.alias:
                # Allow alias=target syntax
                alias, address = args.alias.split('=', 1)

            else:
                # Allow alias target syntax
                alias = args.alias
                address = args.target

            alias = alias.lower()

            if alias in RESERVED_ALIASES:
                log.error("Cannot alias reserved name %r.", alias)
                sys.exit(1)

            if not ADDR_RE.match(address):
                log.error("Address %r is malformed, must be valid 6-char address", address)
                sys.exit(1)

            with Aliases() as aliases:
                aliases.add(alias, address)
                # Adding alias is a silent op

    else:
        with Aliases() as aliases:
            for alias, addr in sorted(aliases.aliases.iteritems()):
                context.output({alias: addr})


def unalias(parser, context, args):
    """Undefines an existing alias"""
    handler = ArgumentHandler(
        usage="snap [global options] config unalias [-h] (<alias> | all)",
        description="Undefine an alias for SNAP node addresses ('all' will remove all)",
    )
    handler.add_argument("alias")
    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to config unalias command: %r", rest)
        sys.exit(1)

    log.debug("Unaliasing: %r", args.alias)

    if args.alias.lower() == 'all':
        with Aliases() as aliases:
            aliases.clear()
    else:
        with Aliases() as aliases:
            aliases.discard(args.alias)
