"""SNAPtoolbelt Help Commands"""

import logging
from argparse import RawDescriptionHelpFormatter

from arghandler import ArgumentHandler

from snapconnect.snap import SnapConstantsHelper
from snapstraction import nvparam

from snaptoolbelt.main import main, DESCRIPTION

log = logging.getLogger(__name__)


def show_help(parser, context, args):
    """Help-related commands"""
    handler = ArgumentHandler(
        usage="snap [global options] help [-h] <command> [options]",
        description=DESCRIPTION,
        formatter_class=RawDescriptionHelpFormatter,
        add_help=False
    )
    if len(args) == 0 or args == ['-h'] or args == ['--help']:
        handler.print_help()
    elif args == ['nvparam']:
        help_nvparam()
    else:
        args.append('--help')
        main(args)


# Convert a help_string from NV_PARAM_HELP_STRINGS into a list of the help info for each bit
def extract_bits_help_from_help_string(help_string):
    """Turn NV_PARAM_HELP_STRINGS into help info"""
    return [help_line.split(' = ')[1]
            for help_line in help_string.split('\n')[1:]]


# Generate a dictionary of bit names to bit help for a given nv param
def bits_help_from_param_name(param_name):
    """Extract bit help for specific NV Param by name"""
    help_string = SnapConstantsHelper.NV_PARAM_HELP_STRINGS[
        nvparam.nvparam_lookup[param_name]['id']
    ]
    bits_help = extract_bits_help_from_help_string(help_string)
    # Make a bitfield help dictionary where the keys are the bit names as defined
    # in bitfields and the values are the bit help strings derived from the help string
    return dict(zip(nvparam.nvparam_lookup[param_name]['bitfields'].iterkeys(),
                    bits_help))

# Generate a list of the names of all nvparams that are bitmasks
BITMASK_PARAMS = [
    param_name
    for (param_name, param_info) in nvparam.nvparam_lookup.iteritems()
    if 'bitfields' in param_info
]

# Dynamically generate help information about the various bitfield nvparams
BITFIELDS_HELP = {
    param_name: bits_help_from_param_name(param_name)
    for param_name in BITMASK_PARAMS
}


def help_nvparam():
    """Print the nvparam help"""
    print "SNAPtoolbelt Help -- Available NV params:\n"

    for param_name, param_info in nvparam.nvparam_lookup.iteritems():
        param_id = param_info['id']
        try:
            param_string = SnapConstantsHelper.NV_PARAM_STRINGS[param_id]
            print '    %-44s %s (id %s)' % (param_name, param_string, param_id)
        except KeyError:
            # Ignore params without a help string
            break

        try:
            for bit_name, bit_id in param_info['bitfields'].iteritems():
                print '        %-40s %s (0x%03X)' % (
                    '.' + bit_name, BITFIELDS_HELP[param_name][bit_name],
                    1 << bit_id)
        except KeyError:
            pass
