"""Output formatters for SNAPtoolbelt"""
import logging
import sys

log = logging.getLogger(__name__)


def dkvp_formatter():
    """Create a Delimited Key-Value Pair formatter"""

    # DKVP - http://johnkerl.org/miller/doc/record-heterogeneity.html
    def output(obj):
        """DKVP output"""

        def dict_output(obj):
            """DKVP for dictionaries"""
            print ",".join(["%s=%s" % (str(k).strip(), str(v).strip()) for k, v
                            in sorted(obj.iteritems())])

        def list_output(obj):
            """DKVP for lists/tuples"""
            for item in obj:
                dict_output(item)

        if isinstance(obj, dict):
            dict_output(obj)
        elif isinstance(obj, list) or isinstance(obj, tuple):
            list_output(obj)
        elif isinstance(obj, basestring):
            print obj

        sys.stdout.flush()

    return output


def json_formatter():
    """Create a JSON formatter"""
    import json

    def output(obj):
        """JSON output"""
        print json.dumps(obj, sort_keys=True)
        sys.stdout.flush()

    return output


def json_pretty_formatter():
    """Create a JSON 'pretty-print' formatter"""
    import json

    def output(obj):
        """JSON output"""
        print json.dumps(obj, sort_keys=True, indent=4, default=str)
        sys.stdout.flush()

    return output


def mk_tabular_formatter(fmt="simple"):
    """Create a tabular formatter"""
    def tabular_formatter():
        """Tabular formatter"""
        from tabulate import tabulate

        def output(obj):
            """tabular output"""

            log.debug("Trying to tabulate: %r", obj)
            table_fmt = fmt

            def dict_output(obj):
                """Tabular for dictionaries"""
                print tabulate(((k, obj[k]) for k in sorted(obj.keys())), tablefmt=table_fmt)

            def list_output(obj):
                """tabular for lists/tuples"""
                print tabulate(obj)

            if isinstance(obj, dict):
                dict_output(obj)
            elif isinstance(obj, list) or isinstance(obj, tuple):
                list_output(obj)
            elif isinstance(obj, basestring):
                print obj
            else:
                print "%s" % obj

            sys.stdout.flush()

        return output

    return tabular_formatter


FORMATS = {
    'dkvp': dkvp_formatter,
    'json': json_formatter,
    'json-pretty': json_pretty_formatter,
    'tab': mk_tabular_formatter("plain"),
}
